/////////////////////////////////////////////////////////////////////////////
// Name:        TAS.h
// Purpose:     Declares global objects and widely used classes
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _TAS_H
#define _TAS_H

#include "stdafx.h"
#include <vector>
#include <wx/string.h>
#include <wx/thread.h>
#include <wx/gdicmn.h>

#include <stdint.h>

#define FORWARD       1
#define FASTFORWARD   2
#define STOP          3
#define BACKWARD      4
#define FASTBACKWARD  5

#define MAX(a, b) ((a)>=(b) ? (a) : (b))
#define MIN(a, b) ((a)<=(b) ? (a) : (b))

/////////////////////////////////////////////////////////////////////
//  Global objects

class TASTileSpace;                                   //  Tile space
extern TASTileSpace*        g_pTileSpace;
extern TASTileSpace*        g_pTileSpaceBackup;

class TAS2HAMTileSpace;
extern TAS2HAMTileSpace*    g_p2HAMTileSpace;

class TASkTAMTileSpace;
extern TASkTAMTileSpace*    g_pkTAMTileSpace;

class TASTileTypeSetEx;                               //  Tileset used
extern TASTileTypeSetEx*    g_pEditorTileset;         //  in the editor

class TASContext;                                     //  Context object
extern TASContext*          g_pContext;               //  for preferences

class TASApp;                                         //  Main application
extern TASApp*              g_pTASApp;                //  object

extern wxMutex*             g_pMutexTileSpaceGuard;   // Mutex to guard tile space
                                                      // from concurrent access by
                                                      // main UI thread and iterator

class wxPrintData;                                    // Global print data, to
extern wxPrintData*         g_printData;              // remember settings during
                                                      // the session

class wxPageSetupData;                                // Global page setup data
extern wxPageSetupData*     g_pageSetupData;

//
/////////////////////////////////////////////////////////////////////

#define TAS_COORD_MAX       INT_MAX
#define TAS_COORD_MIN       (INT_MIN + 1)
#define TAS_COORD_INVALID   INT_MIN

#define TAS_COORDS_OFFSETS

class TASCoords
{
public:
  TASCoords() { Invalidate(); };
  TASCoords(int x_in, int y_in, int z_in) { x = x_in; y = y_in; z = z_in; };
  TASCoords(const TASCoords& cIn) { x = cIn.x; y = cIn.y; z = cIn.z; };
  TASCoords(TASCoords* pC) { x = pC->x; y = pC->y; z = pC->z; };
  TASCoords(wxString& sIn) { SetFromString(sIn); }
  TASCoords& operator=(const TASCoords& tc) { x = tc.x; y = tc.y; z = tc.z; return *this; };
  TASCoords& operator=(const wxPoint& pt) { x = pt.x; y = pt.y; z = 0; return *this; };
  TASCoords operator+(const TASCoords& tc) const { return TASCoords(x+tc.x, y+tc.y, z+tc.z); };
  TASCoords operator+(const int dir);
  inline bool operator<(const TASCoords& tc) { return (z < tc.z || (z == tc.z && y < tc.y) || (z == tc.z && y == tc.y && x < tc.x)); };
  inline bool operator<(const TASCoords& tc) const { return (z < tc.z || (z == tc.z && y < tc.y) || (z == tc.z && y == tc.y && x < tc.x)); };
  inline bool operator==(const TASCoords& tc) { return (x == tc.x && y == tc.y && z == tc.z); };
  inline bool operator!=(const TASCoords& tc) { return !(operator==(tc)); };
  inline bool operator==(const wxPoint& pt) const { return (x == pt.x && y == pt.y && z == 0); };
  inline bool operator!=(const wxPoint& pt) const { return !(operator==(pt)); };
  bool SetFromString(wxString& sIn);
  wxString ToString() const;
  inline bool IsValid() const { return ((x!=TAS_COORD_INVALID)&&(y!=TAS_COORD_INVALID)&&(z!=TAS_COORD_INVALID)); };
  inline void Invalidate() { x = TAS_COORD_INVALID; y = TAS_COORD_INVALID; z = TAS_COORD_INVALID; };
  friend bool operator==(const TASCoords& lhs, const TASCoords& rhs);

  // HashType must be a 64-bit type
  typedef uint64_t HashType;
  // CoordType must be a 32-bit integer
  typedef int32_t CoordType;

  // (Michael) Only hashes coords well when  -(2^20) <= x, y, z <= (2^20)
  HashType hash() const {
	  HashType hash(0);
	  // Take care of negatives
	  if(x < 0)
		  hash |= 0x1;
	  if(y < 0)
		  hash |= 0x2;
	  if(z < 0)
		  hash |= 0x4;

	  hash <<= 20;
	  hash |= (x & 0xFFFFF);
	  hash <<= 20;
	  hash |= (y & 0xFFFFF);
	  hash <<= 20;
	  hash |= (z & 0xFFFFF);

	  return hash;
  }

  CoordType x;
  CoordType y;
  CoordType z;
};

class TASSpace
{
public:
  TASSpace() {};

  virtual int GetXDim() = 0;
  virtual int GetYDim() = 0;
  virtual int GetZDim() = 0;
  virtual size_t Index(const TASCoords& coords) = 0;
  virtual TASCoords CoordsFromIndex(size_t ind) = 0;
};

#endif /* _TAS_H */

