/////////////////////////////////////////////////////////////////////////////
// Name:        TASCuts.h
// Purpose:     Defines classes used for calculating minimum cuts of assemblies
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2010, 2011 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#include "TASCuts.h"
#include "TASApp.h"
#include "TASTiles.h"

#ifdef __WXMSW__
    #include <wx/msw/msvcrt.h>      // redefines the new() operator 
#endif 

TASCoords TASCoords::operator+(const int dir)
{
  TASCoords c(x,y,z);
  if (dir == TAS_NORTH) c.y += 1;
  else if (dir == TAS_EAST) c.x += 1;
  else if (dir == TAS_SOUTH) c.y -= 1;
  else if (dir == TAS_WEST) c.x -= 1;
  else if (dir == TAS_UP) c.z += 1;
  else if (dir == TAS_DOWN) c.z -= 1;
  return c;
}

TASCutSet::TASCutSet(const TASCutSet& cutSetIn)
{
  m_locs      = cutSetIn.m_locs;
  m_strength  = cutSetIn.m_strength;
}

wxString TASCutSet::ToString()
{
  wxString sRet;
  for (std::set<TASCoords>::iterator it = m_locs.begin(); it != m_locs.end(); it++) {
    TASCoords c((const TASCoords)*it);
    sRet = sRet + c.ToString() + wxT(" ");
  }
  return sRet;
}

void TASCutSets::CheckCutsRec(TASCoords coords, std::set<TASCoords> setLiveNbrs, std::set<TASCoords> setDeadNbrs, TASCutSet csCurrCut, int threshold, std::vector<TASCutSet>* pSetCuts)
{  
  //  Add this neighbor to the cut and remove it from the live set
  csCurrCut.m_locs.insert(coords);
//  std::set<TASCoords>::iterator it = setLiveNbrs.find(coords);
//  setLiveNbrs.erase(it);

  if (csCurrCut.m_locs.size() == (m_pTileSpace->m_numTiles+1)) {    //  Get outta here - we've got the entire assembly
    return;
  }

  //  Correctly add/subtract strenghts of bindings of this tile to/from the current cut
  std::vector<TASCoords> vectCoordsNewNbrs;
  for (int dir=TAS_NORTH; dir<=m_pTileSpace->GetNumNbrs(); dir++) {
    TASCoords nbr2 = coords + dir;
    TASTileType* nbr2TType = m_pTileSpace->GetTileType(nbr2);
    if (nbr2TType) {
      TASTileType* nbrTType = m_pTileSpace->GetTileType(coords);
      int str = m_pTileSpace->m_tileset.GetGlueInteractionStrength(nbrTType->m_sides[dir], nbr2TType->m_sides[TASOppositeDirection[dir]]);

      if (csCurrCut.m_locs.find(nbr2) != csCurrCut.m_locs.end()) {    //  This edge touches part of the current cut
        csCurrCut.m_strength -= str;
      } else {
        csCurrCut.m_strength += str;                                  //  This edge is bound outside of the cut
        if (str > 0) {
          //  If this neighbor isn't already in the live set and is also not in the dead set, add it to the live set
          if ((setLiveNbrs.find(nbr2) == setLiveNbrs.end()) && (setDeadNbrs.find(nbr2) == setDeadNbrs.end())) {
            setLiveNbrs.insert(nbr2);
            vectCoordsNewNbrs.push_back(nbr2);
          }
        }
      }
    }
  }

/*
wxString sMsg = wxString::Format(wxT("csCurrCut (%d): "), csCurrCut.m_strength);
for (std::set<TASCoords>::iterator it = csCurrCut.m_locs.begin(); it != csCurrCut.m_locs.end(); it++) {
  sMsg += it->ToString();
  sMsg += " ";
}
sMsg += wxT("\n");
g_pTASApp->SetSimDebugText(sMsg, true, false);

wxString sMsg3 = wxString::Format(wxT("\tliveNbrs: "));
for (std::set<TASCoords>::iterator it = setLiveNbrs.begin(); it != setLiveNbrs.end(); it++) {
  sMsg3 += it->ToString();
  sMsg3 += " ";
}
sMsg3 += wxT("\n");
g_pTASApp->SetSimDebugText(sMsg3, true, false);
*/

  //  If the current cut is <= str of those in pSetCuts and below threshold, make a copy of it in pSetCuts
  if (csCurrCut.m_strength <= threshold) {
    if ((pSetCuts->size() > 0) && (csCurrCut.m_strength < pSetCuts->at(0).m_strength)) {
      pSetCuts->clear();
    }
    if ((pSetCuts->size() == 0) || (csCurrCut.m_strength == pSetCuts->at(0).m_strength)) {
      pSetCuts->push_back(csCurrCut);
    }
  }

//MJP TODO: get rid of vectCoordsNewNbrs
  //  Recursively call CheckCutsRec once for each neighbor that we just added to the live set
//MJP TODO: this would be a good place to inject randomness, but it's not very efficient with a set
//  int startPos = g_pTASApp->RandInRange((int)setLiveNbrs.size());
//  for (size_t i=0; i<vectCoordsNewNbrs.size(); i++) {
//    TASCoords currNbr = vectCoordsNewNbrs[(i+startPos)%vectCoordsNewNbrs.size()];
  std::set<TASCoords> setLiveNbrsToPass(setLiveNbrs);
  for (std::set<TASCoords>::iterator it = setLiveNbrs.begin(); it != setLiveNbrs.end(); it++) {

    setLiveNbrsToPass.erase(*it);

    //  Recurse with this neighbor added to the cut
    CheckCutsRec(*it, setLiveNbrsToPass, setDeadNbrs, csCurrCut, threshold, pSetCuts);

    //  Now remove this neighbor and add it to the dead list, then recurse with the other neighbors
//    std::set<TASCoords>::iterator it = setLiveNbrs.find(currNbr);

/*
wxString sMsg2 = wxT("Adding location ");
sMsg2 += it->ToString();
sMsg2 += wxT(" to dead list\n");
g_pTASApp->SetSimDebugText(sMsg2, true, false);
*/
    setDeadNbrs.insert(*it);
  }

  return;
}

void TASCutSets::CheckCuts(TASCoords loc)
{
  clear();    //  Clear out the previously calculated cuts

  std::vector<TASTileBinding> vectBindingSides;
  int min = m_pTileSpace->GetTileBindings(loc, &vectBindingSides);

  if (min < 0) {
    std::set<TASCoords> setLiveNbrs;
    std::set<TASCoords> setDeadNbrs;
    TASCutSet csCurrCut;
    std::vector<TASCutSet> vectCS;

    for (size_t i=0; i<vectBindingSides.size(); i++) {
      if (vectBindingSides[i].m_strength < 0) {
        CheckCutsRec(loc, setLiveNbrs, setDeadNbrs, csCurrCut, m_pTileSpace->GetTemp()-1, &vectCS);

        TASCoords loc2(loc);
        CheckCutsRec(loc2 + i, setLiveNbrs, setDeadNbrs, csCurrCut, m_pTileSpace->GetTemp()-1, &vectCS);
      }
    }     

    if (vectCS.size() > 0) {
      wxString sMsg = wxString::Format(wxT("Step %d: Tile addition to location %s caused a minimum cut of strength %d.\n"), m_pTileSpace->m_epoch, loc.ToString().c_str(), vectCS[0].m_strength);
      g_pTASApp->SetSimOutputText(sMsg, true, false);

      sMsg = wxString::Format(wxT("Step %d: min cut = %d, number sets with that min cut = %d.\n"), m_pTileSpace->m_epoch, vectCS[0].m_strength, vectCS.size());
      for (size_t i=0; i<vectCS.size(); i++) {
        sMsg += wxString::Format(wxT("\tContents of set %d: "), i);
        for (std::set<TASCoords>::iterator it = vectCS[i].m_locs.begin(); it != vectCS[i].m_locs.end(); it++) {
	  TASCoords c((const TASCoords)*it);
          sMsg += c.ToString();
          sMsg += wxT(" ");
        }
        sMsg += wxT("\n");
      }
      g_pTASApp->SetSimDebugText(sMsg, true, false);

      //  Decide which cut to get rid of.  For now, that means the first one we find which doesn't contain
      //  part of the seed.
      for (size_t i=0; i<vectCS.size(); i++) {
        bool bHasSeed = false;
//MJP TODO: this could be done more efficiently by iterating over the seed and searching the set
        for (std::set<TASCoords>::iterator it = vectCS[i].m_locs.begin(); it != vectCS[i].m_locs.end(); it++) {
          if (m_pTileSpace->IsInSeed((TASCoords&)*it)) {
            bHasSeed = true;
            break;
          }
        }
        if (!bHasSeed) {        //  Flag this cut to be discarded
          push_back(vectCS[i]);
        }
      }
    }
  }
}

void TASCutSets::DumpToFile(wxString sFile)
{
  wxFFile f(sFile, _T("w"));
  
  for (size_t i=0; i<size(); i++) {
    TASCutSet* pCS = &this->at(i);
    wxString str;
    str = str.Format(wxT("%d: "), pCS->m_strength);
    for (std::set<TASCoords>::iterator it = pCS->m_locs.begin(); it != pCS->m_locs.end(); it++) {
      TASCoords c((const TASCoords)*it);
      str += c.ToString();
      str += wxT(" ");
    }
    str += wxT("\n");
    f.Write(str);
  }
  f.Close();
}